/*
 *	eZ80 Processor Module
 *	Brandon Wilson
 *	2015
 */

#include "ez80.hpp"
#include <entry.hpp>
#include <diskio.hpp>

static const char *RegNames[] =
  {
    "b",  "c",  "d", "e", "h", "l", "", "a",
    "bc", "de", "hl", "af", "sp", "ix", "iy", "af'",
    "r",  "i",  "f", "xl", "xh", "yl", "yh",
    /*"w",  "lw",*/  "ixl", "ixh", /*"dsr", "xsr",*/ "iyl",
    "iyh", /*"ysr", "sr", "ib", "iw", "xm", "lck",*/
    "bc'", "de'", "hl'","ix'","iy'",
    "b'",  "c'",  "d'", "e'", "h'", "l'", /*"m'",*/ "a'",
	"mb",
    "cs","ds","T"
  };

//TASM assembler definition for Z80
static ushort tasmz80_bads[] = { I5_rst, Z80_srr, 0 };
static asm_t tasmz80 = {
  AS_COLON | AS_N2CHR | AS_ASCIIC,
  0,
  "Table Driven Assembler (TASM)",
  0,
  NULL,
  tasmz80_bads,
  ".org",
  ".end",
  ";",          // comment string
  '\'',         // string delimiter
  '\'',         // char delimiter
  "\\\"'",      // special symbols in char and string constants
  ".ascii",     // ascii string directive
  ".db",        // byte directive
  ".dw",        // word directive
  ".dl",         // dword  (4 bytes)
  NULL,         // qword  (8 bytes)
  NULL,         // oword  (16 bytes)
  NULL,         // float  (4 bytes)
  NULL,         // double (8 bytes)
  NULL,         // tbyte  (10/12 bytes)
  NULL,         // packed decimal real
  NULL,         // arrays (#h,#d,#v,#s(...)
  ".ds %s",  // uninitted arrays
  ".equ",         // equ
  NULL,         // seg prefix
  NULL,         // checkarg_preline
  NULL,         // checkarg_atomprefix
  NULL,         // checkarg_operations
  NULL,         // XlatAsciiOutput
  "$",         // curip
  NULL,         // func_header
  NULL,         // func_footer
  NULL,         // public
  NULL,         // weak
  NULL,         // extrn
  NULL,         // comm
  NULL,         // get_type_name
  NULL,         // align
  ' ', ' ',     // lbrace, rbrace
  "%",			// mod
  "&",			// and
  "|",			// or
  "^",			// xor
  "~",			// not
  "<<",			// shl
  ">>",			// shr
  NULL,			// sizeof
};

static asm_t *Z80asms[]     = { &tasmz80, NULL };
static netnode helper;
char device[MAXSTR] = "";
static size_t numports;
static ioport_t *ports;

#include "../iocommon.cpp"
const char *z80_find_ioport(uval_t port)
{
  const ioport_t *p = find_ioport(ports, numports, port);
  return p ? p->name : NULL;
}

const char *idaapi set_idp_options(const char *keyword,int /*value_type*/,const void * /*value*/)
{
  if ( keyword != NULL ) return IDPOPT_BADKEY;
  char cfgfile[QMAXFILE];
  get_cfg_filename(cfgfile, sizeof(cfgfile));
  if ( choose_ioport_device(cfgfile, device, sizeof(device), parse_area_line0) )
    set_device_name(device, IORESP_NONE);
  return IDPOPT_OK;
}

static int notify(processor_t::idp_notify msgid, ...) { // Various messages:
  va_list va;
  va_start(va, msgid);

// A well behaving processor module should call invoke_callbacks()
// in his notify() function. If this function returns 0, then
// the processor module should process the notification itself
// Otherwise the code should be returned to the caller:

  int code = invoke_callbacks(HT_IDP, msgid, va);
  if ( code ) return code;

  switch ( msgid )
  {
    case processor_t::init:
      helper.create("$ z80");
      break;

    case processor_t::newprc:
      {
        ph.assemblers = Z80asms;
        {
          char buf[MAXSTR];
          if ( helper.supval(-1, buf, sizeof(buf)) > 0 )
            set_device_name(buf, IORESP_NONE);
        }
      }
      break;

    case processor_t::newfile:
      if ( strcmp(inf.procName, "z180") == 0 )
      {
        char cfgfile[QMAXFILE];
        get_cfg_filename(cfgfile, sizeof(cfgfile));
        if ( choose_ioport_device(cfgfile, device, sizeof(device), parse_area_line0) )
          set_device_name(device, IORESP_AREA);
      }
      break;

    default:
      break;
  }
  va_end(va);

  return(1);
}

static bool idaapi can_have_type(op_t &x) //returns 1 - operand can have
{
  switch ( x.type )
  {
    case o_void:        // No Operand
    case o_reg:         // General Register
    case o_phrase:      // Base Reg + Index Reg
    case o_cond:        // FPP register
      return 0;
  }
  return 1;
}

static const char *shnames[] = { "eZ80", NULL };
static const char *lnames[] = { "Zilog eZ80", NULL };

static uchar retcode_1[] = { 0xC9 };
static uchar retcode_2[] = { 0xED, 0x45 };
static uchar retcode_3[] = { 0xED, 0x4D };

static bytes_t retcodes[] =
{
 { sizeof(retcode_1), retcode_1 },
 { sizeof(retcode_2), retcode_2 },
 { sizeof(retcode_3), retcode_3 },
 { 0, NULL }
};

// Zilog eZ80 processor definition
processor_t LPH =
{
  IDP_INTERFACE_VERSION,
  PLFM_Z80,                     // id
  PRN_HEX|PR_SEGTRANS|PR_SEGS|PR_USE32|PR_DEFSEG32,
  8,                            // 8 bits in a byte for code segments
  8,                            // 8 bits in a byte for other segments


  shnames,
  lnames,

  Z80asms,

  notify,

  ez80_header,
  ez80_footer,

  ez80_segstart,
  std_gen_segm_footer,

  NULL,                         // assumes

  ana,
  emu,

  out,
  outop,
  intel_data,
  NULL,                         // int  (*cmp_opnd)(op_t &op1,op_t &op2);
                                // returns 1 - equal operands
  can_have_type,                // returns 1 - operand can have
                                // a user-defined type

  R_vt+1,                       // number of registers
  RegNames,
  NULL,

  0,
  NULL,
  NULL,
  NULL,
//
//      There will be 2 virtual registers: code segment register
//                                         data segment register
  R_vcs,R_vt,                  // first, last
  4,                           // size of a segment register
  R_vcs,R_vds,                 // CS,DS

  NULL,                        //No known code start sequences
  retcodes,                    //'Return' instruction codes

  0,Z80_last,
  Instructions,
  NULL,                 // int  (*is_far_jump)(int icode);
  NULL,                 // Translation function for offsets
  3,                    // int tbyte_size;  -- doesn't exist
  NULL,                 // int (*realcvt)(void *m, ushort *e, ushort swt);
  { 0, 0, 0, 0 },       // char real_width[4];
                        // number of symbols after decimal point
                        // 2byte float (0-does not exist)
                        // normal float
                        // normal double
                        // long double
  NULL,                 // int (*is_switch)(switch_info_t *si);
  NULL,                 // long (*gen_map_file)(FILE *fp);
  NULL,                 // ea_t (*extract_address)(ea_t ea,const char *string,int x);
  NULL,                 // int (*is_sp_based)(op_t &x); -- always, so leave it NULL
  NULL,                 // int (*create_func_frame)(func_t *pfn);
  NULL,                 // int (*get_frame_retsize(func_t *pfn)
  NULL,                 // void (*gen_stkvar_def)(char *buf,const member_t *mptr,sval_t v);
  gen_spcdef,           // Generate text representation of an item in a special segment
  I5_ret,               // Icode of return instruction. It is ok to give any of possible return instructions
  set_idp_options,      // const char *(*set_idp_options)(const char *keyword,int value_type,const void *value);
  NULL,                 // int (*is_align_insn)(ea_t ea);
  NULL,                 // mvm_t *mvm;
};
